﻿// TODO: Delete this code when removing V1
var glob = require("glob"),
    Promise = require('promise'),
    path = require('path'),
    fs = require('fs'),
    fileAndFolderUtils = require('../file-and-folder-utils'),
    fileSignature = require('./file-signature.js'),
    methodKeyCreator = require('./method-key.js'),
    branchKeyCreator = require('./branch-key.js'),
    uuid = require('uuid'),
    utils = require('../utils'),
    consoleFileLogger = require("../sl-logger.js").consoleFileLogger,
    ActualFile = require('../file-system/actual-file'),
    InMemoryFile = require('../file-system/in-memory-file'),
    SlJsInfra = require('sl-js-infra').SlJsInfra,
    SlIgnoreWrapper = SlJsInfra.SlIgnoreWrapper;

function javascriptHandler(handlerParams, logger) {
    //V1 protocol
    var log = createLogger(logger, "javascriptHandler");
    log.info("Inside javascriptHandler");
    var opts = handlerParams.buildArguments;
    var slIgnoreScan = createSlIgnore(opts, SlIgnoreWrapper.SCAN_IGNORE_FILE_NAME, logger, log);
    var slIgnoreGenerated = createSlIgnore(opts, ".slignore.generated", logger, log);
    var finalOutput = handlerParams.finalOutput;
    var p = new Promise(function (resolve, reject) {
        finalOutput.files = finalOutput.files || [];
        var files = [];
        var filesLookup = {};
        handlerParams.processedFiles = [];
        var erroredFiles = [];
        populateFilesArray(handlerParams, opts, filesLookup, files);
        files.forEach(function (filename, fileIdx) {
            handleFileData(handlerParams, filename,  fileIdx, files, opts, logger, slIgnoreScan, slIgnoreGenerated, finalOutput, erroredFiles, log);
        });
        var result = handleErroredFiles(erroredFiles, log);
        return resolve(result);
    });
    return p;
}

function javascriptHandlerV2(handlerParams, logger) {
    return javascriptHandler(handlerParams, logger).then(function () {
        var p = new Promise(function (resolve, reject) {
            var log = createLogger(logger, "javascriptHandlerV2");
            var opts = handlerParams.buildArguments;
            var finalOutput = handlerParams.finalOutput;
            var fileToCommitsMap = handlerParams.fileToCommitsMap;
            var unsortedFiles = finalOutput.files;
            var filesLookup = {};
            var filesArray = [];
            finalOutput.files = filesArray;
            finalOutput.version = "2.0";
            unsortedFiles.forEach(function (file) {
                if (!file) return;
                if (file.methods) {
                    file.methods.forEach(function (method) {
                        handleMethodInFile(method, file, filesLookup, opts, fileToCommitsMap, filesArray, log);
                    })
                }

                if (file.branches) {
                    file.branches.forEach(function (branch) {
                        handleBranchInFile(branch, file, filesLookup, opts, fileToCommitsMap, filesArray, log);
                    });
                }
            });
            return resolve(true);
        });
        return p;
    });
}

function generateNewMappingForFile(fileObject, isBranchCoverage, es6Modules, prefixesOfExcludedFiles, useBabylon, logger) {
    var filename = fileObject.getFilename();
    var fullname = fileObject.getFullFilename();
    filename = utils.adjustPathSlashes(filename);
    var fileSig = {
        filename: filename,
        methods: []
    };

    var oldContent = fileObject.getContent();
    var sourceMaps = fileObject.readSourceMaps(logger);
    var sig = fileSignature.generateFileSignature(fullname, filename, oldContent, sourceMaps, isBranchCoverage, {
        es6Modules: es6Modules,
        useBabylon: useBabylon,
        prefixesOfExcludedFiles: prefixesOfExcludedFiles
    }, logger);

    if (!sig || sig.error) {
        return null;
    }
    fileSig.hash = sig.module.hash;
    fileSig.hasSourceMaps = !!sourceMaps;

    var mapping = {};
    sig.methods.forEach(function (m) {
        var wholeKey = [methodKeyCreator(m.start, m.name, m.isAnonymous), m.hash, new Date().valueOf()].join('|');
        var newId = uuid(wholeKey);
        fileSig.methods.push({
            id: newId,
            isAnonymous: m.isAnonymous,
            hash: m.hash,
            sigHash: m.sigHash,
            type: m.type,
            position: m.start,
            endPosition: m.end,
            name: m.name,
            uniqueName: m.uniqueName,
            filename: m.originalFilename,
            originalFilename: m.filename,
            originalStart: m.originalStart,
            originalEnd: m.originalEnd,
            originalName: m.originalName,
            branches: m.branches
        })
    });
    return fileSig;
}

function mapSingleFileInMemory(singleFileParams, es6Modules, prefixesOfExcludedFiles, useBabylon, logger) {
    var useBranchCoverage = singleFileParams.usebranchcoverage;
    var fileObject = new InMemoryFile(singleFileParams.filename, singleFileParams.originalFilename, singleFileParams.buffer);
    var fileSig = generateNewMappingForFile(fileObject, useBranchCoverage, es6Modules, prefixesOfExcludedFiles, useBabylon, logger);
    return fileSig;
}

function populateFilesArray(handlerParams, opts, filesLookup, files) {
    handlerParams.buildArguments.includedFiles.forEach(function (inclusionPattern) {
        var matchedFiles = glob.sync(inclusionPattern, {
            cwd: opts.workspacepath,
            ignore: handlerParams.buildArguments.excludedFiles
        });
        matchedFiles.forEach(function (matchedFile) {
            if (!filesLookup[matchedFile]) {
                filesLookup[matchedFile] = true;
                files.push(matchedFile);
            }
        });
    });
}

function createLogger(logger, methodName) {
    var log = logger ? logger.child({
        methodName: methodName
    }) : consoleFileLogger(methodName);
    return log;
}

function createFileSignature(handlerParams, opts, filename, fullFilename, logger) {
    var useBranchCoverage = handlerParams.cfg.useBranchCoverage || opts.usebranchcoverage;
    var fileObject = new ActualFile(filename, fullFilename);
    var fileSig = generateNewMappingForFile(fileObject, useBranchCoverage, opts.es6Modules || opts["es-modules"], opts.prefixesOfExcludedFiles, opts.useBabylon, logger);
    return fileSig;
}

function shouldSkipFile(fullFilename, filename, slIgnore, log, fileIdx, files) {
    var filePath = fs.statSync(fullFilename);
    if (!filePath.isFile())
        return true;

    // If the file is filtered by the .slignore file:
    if (slIgnore.denies(filename)) {
        log.debug('Excluded [' + (fileIdx + 1) + '/' + files.length + '] ' + fullFilename);
        return true; // Ignored by .slignore, skip it.
    }
    return false;
}

function handleGeneratedMethod(slIgnore, method, includedMethods) {
    if (slIgnore.accepts(method.originalFilename)) {
        includedMethods.push(method);
    } else {
        console.log('Rejecting method from file: ' + method.originalFilename);
    }
}

function getIncludedMethods(fileSig, slIgnore) {
    if (!fileSig.methods)
        return [];
    var includedMethods = [];
    fileSig.methods.forEach(function (method) {
        //If source maps are used, the methods in fileSig may originate from excluded files
        if (method.originalFilename) {
            handleGeneratedMethod(slIgnore, method, includedMethods);
        } else {
            includedMethods.push(method);
        }
    });
    return includedMethods;
}

function getIncludedBranches(fileSig, slIgnore) {
    if (!fileSig.branches)
        return []
    var includedBranches = [];
    fileSig.branches.forEach(function (branch) {
        if (branch.filename) {
            if (slIgnore.accepts(branch.filename)) {
                includedBranches.push(branch);
                delete branch.filename;
            }
            else {
                //console.log('Rejecting branch from file: '+branch.filename);
            }
        } else {
            includedBranches.push(branch);
        }
    });

    return includedBranches;
}

function handleFileData(handlerParams, filename, fileIdx, files, opts,
                        logger, slIgnoreScan, slIgnoreGenerated, finalOutput,
                        erroredFiles, log) {
    var fullFilename = path.resolve(opts.workspacepath, filename);
    var shouldSkipCurrFile = shouldSkipFile(fullFilename, filename, slIgnoreScan, log, fileIdx, files);
    if (shouldSkipCurrFile)
        return;
    handlerParams.processedFiles.push(filename);
    log.info('Processing [' + (fileIdx + 1) + '/' + files.length + '] ' + fullFilename);
    var fileSig = createFileSignature(handlerParams, opts, filename, fullFilename, logger);
    if (fileSig) {
        fileSig.methods = getIncludedMethods(fileSig, slIgnoreGenerated);
        fileSig.branches = getIncludedBranches(fileSig, slIgnoreGenerated);
        finalOutput.files.push(fileSig);
    }
    else {
        erroredFiles.push(filename);
    }
}

function handleErroredFiles(erroredFiles, log) {
    var result = true;
    if (erroredFiles.length) {
        log.warn(
            "The following files failed to be parsed and will not be sent to the server:\n" +
            erroredFiles.join('\n') +
            "\n\t(Hint: if the file use the es6 module syntax, pass the \"--es6Modules\" switch)");
        result = false; //Don't fail the build scanning process if some files failed to be parsed
    }
    return result;
}

function createNewFile(filename) {
    var newFile = {
        logicalPath: filename,
        physicalPath: filename,
        methods: [],
        lines: [],
        branches: []
    };
    return newFile;
}

function logCommitsArrayNotFound(opts, fileToCommitsMap, log) {
    var fieldsMap = {
        opts: !opts,
        sendContributors: opts && !opts.sendContributors,
        fileToCommitsMap: !fileToCommitsMap
    }
    var missing = utils.getMissingProperties(fieldsMap);
    log.error("Cannot add commits indexes, missing fields: " + missing);
}

function shouldAddCommitsArray(opts, fileToCommitsMap) {
    return opts && opts.sendContributors && fileToCommitsMap;
}

function handleNewFile(filename, opts, fileToCommitsMap, log, filesLookup, filesArray) {
    var newFile = createNewFile(filename);
    var addCommitsArray = shouldAddCommitsArray(opts, fileToCommitsMap)
    if (addCommitsArray) {
        newFile.commitIndexes = fileToCommitsMap[filename] || [];
    } else {
        logCommitsArrayNotFound(opts, fileToCommitsMap, log);
    }
    filesLookup[filename] = newFile;
    filesArray.push(newFile);
}

function getCreateFile(filename, filesLookup, opts, fileToCommitsMap, filesArray, log) {
    if (!filesLookup[filename]) {
        handleNewFile(filename, opts, fileToCommitsMap, log, filesLookup, filesArray);
    }
    return filesLookup[filename];
}

function convertMethod(trueSrcFilename, method) {
    var convertedMethod = {
        uniqueId: [trueSrcFilename, method.position.join(',')].join('@'),
        displayName: method.name,
        position: method.position,
        endPosition: method.endPosition,
        hash: method.hash,
        sigHash: method.sigHash,
        meta: {} //Filled-in below
    };
    if (method.isAnonymous) {
        convertedMethod.meta.anonymous = true;
    }
    if (method.type) {
        convertedMethod.meta.type = method.type;
    }
    return convertedMethod;
}

function extractBranches(branch, method, file) {
    if (!branch) return;
//TODO: Hack! Currently the branches which we collect are inside method.
    //When we'll support global branches, we should get the "original file name"
    //in the same time we create the branch data object.
    branch.originalFilename = method.originalFilename;
    file.branches.push(branch);
}

function handleMethodInFile(method, file, filesLookup, opts, fileToCommitsMap, filesArray, log) {
    if (!method) return;
    var trueSrcFilename = method.originalFilename || file.filename;
    if (file.hasSourceMaps && trueSrcFilename == file.filename) { // Skip method that is not mapped to any "true" source file. It's compiler generated
        return;
    }
    var methodFile = getCreateFile(trueSrcFilename, filesLookup, opts, fileToCommitsMap, filesArray, log);
    var convertedMethod = convertMethod(trueSrcFilename, method);
    methodFile.methods.push(convertedMethod);
    file.branches = file.branches || [];
    if (method.branches) {
        method.branches.forEach(function (branch) {
            return extractBranches(branch, method, file);
        });
    }
}

function createBranchData(uniqueId, branch) {
    var branchData = {
        uniqueId: uniqueId,
        filename: branch.filename,
        type: branch.type,
        position: branch.position,
        endPosition: branch.endPosition,
        hash: branch.hash,
        index: branch.index
    }
    return branchData;
}

function handleBranchInFile(branch, file, filesLookup, opts, fileToCommitsMap, filesArray, log) {
    if (!branch) return;
    var trueSrcFilename = branch.originalFilename || file.filename;
    if (file.hasSourceMaps && trueSrcFilename == file.filename) { // Skip method that is not mapped to any "true" source file. It's compiler generated
        return;
    }
    var branchFile = getCreateFile(trueSrcFilename, filesLookup, opts, fileToCommitsMap, filesArray, log);
    var uniqueId = branchKeyCreator(branch.originalFilename, branch.parentPosition, branch.index);
    var branchData = createBranchData(uniqueId, branch);
    branchFile.branches.push(branchData);
}

function createSlIgnore(opts, filename, logger, log) {
    var slIgnoreScanFilePath = fileAndFolderUtils.findFileRecursively(filename, opts.workspacepath, !!opts.instrumentForBrowsers, logger);
    var slIgnoreScan = new SlIgnoreWrapper(slIgnoreScanFilePath, opts.outputpath, filename, log);
    return slIgnoreScan;
}

module.exports.javascriptHandler = javascriptHandler;
module.exports.javascriptHandlerV2 = javascriptHandlerV2;
module.exports.mapSingleFileInMemory = mapSingleFileInMemory;